/*
 * @(#)CrossKeyboard.java   1.0 07/01/2000
 */

import java.io.*;
import java.text.*;

/**
 * CrossKeyboard.java - Keyboard driver of Cross Process for continous beams.
 *
 * <p>
 * Description:
 * <p>
 * ===============================================================
 * <p>
 * This file contains driver routines for solving moments of 
 * continuous beams by the Cross Process.  For more details of 
 * this process, refer to class notes of course CIV1127-Analise 
 * de Estruturas II, Dept. of Civil Engineering, PUC-Rio.
 *
 * <p>
 * ===============================================================
 *
 * @version    1.0 07/01/2000
 * @author     Luiz Fernando Martha
 * @author     Andre Cahn
 *
 */
public class CrossKeyboard
{

 /*
 ** ---------------------------------------------------------------
 ** Class variables:
 */

 /** Number of members. */
 static int n_member;

 /** Left end support condition. */
 static int sup_init = 0;

 /** Right end support condition. */
 static int sup_end = 0;

 /** Tolerance for moment balancing. */
 static double mom_tol = 0.01;

 /** Main array of member length */
 static double len[];

 /** Main array of member uniform load */
 static double q[];

 /** Main array of member left moment */
 static double ml[];

 /** Main array of member right moment */
 static double mr[];

/** Main array of node rotation */
 static double rot[];


 /*
 ** ---------------------------------------------------------------
 ** Class methods: 
 */

 /*===========================  getData  ===========================*/

 static boolean getData()
 {
  try
  {
   BufferedReader in = new BufferedReader( new InputStreamReader(System.in) );
   String line;

   System.out.print( "\nEnter number of members: " );
   line = in.readLine();
   if( line.length() == 0 ) return( false );
   n_member = Integer.parseInt( line );
   if( n_member < 2 )
   {
    System.out.println( "Number of members has to be at least 2." );
    return( false );
   }

   System.out.print( "Enter initial support condition (0=free, 1=fix): " );
   line = in.readLine();
   if( line.length() == 0 ) return( false );
   sup_init = Integer.parseInt( line );
   if( sup_init != 0 )
    sup_init = 1;

   System.out.print( "Enter end support condition (0=free, 1=fix): " );
   line = in.readLine();
   if( line.length() == 0 ) return( false );
   sup_end = Integer.parseInt( line );
   if( sup_end != 0 )
    sup_end = 1;

   len = new double[n_member];
   q = new double[n_member];
   for( int i=0; i<n_member; i++ )
   {
    System.out.print( "Enter member length: " );
    line = in.readLine();
    if( line.length() == 0 ) return( false );
    len[i] = Double.parseDouble( line );
    System.out.print( "Enter member load: " );
    line = in.readLine();
    if( line.length() == 0 ) return( false );
    q[i] = Double.parseDouble( line );
   }
  }
  catch( Exception e )
  {
   System.out.println( e );
   return( false );
  }
  return( true );
 }

 /*========================= printResults ==========================*/

 static void printResults()
 {
  DecimalFormat form_mom;
  DecimalFormat form_rot;

  try {
   form_mom = new DecimalFormat( "##0.00" );
   form_rot = new DecimalFormat( "##0.000" );
  } 
  catch( IllegalArgumentException e ) {
   return;
  }

  System.out.println( "\n" );
  for( int i=0; i<n_member; i++ )
  {
   System.out.println( "Member " + i +
                       ": Left Moment = " + form_mom.format( ml[i] ) +
                       "  Right Moment = " + form_mom.format( mr[i] ) );
  }
  for( int i=0; i<n_member-1; i++ )
  {
   System.out.println( "Node " + i +
   	                   ": Rotation = " + form_rot.format(rot[i]*1000.0) +
                       " E-03" );
  }
 }

 /*============================  main  =============================*/

 public static void main(String[] args)
 {
  CrossSolver solver;

  if( ! getData() )
   return;
  solver = new CrossSolver( n_member, sup_init, sup_end, mom_tol, len, q );

  n_member = solver.getNumMembers();
 
  ml  = new double[n_member];
  mr  = new double[n_member];
  rot = new double[n_member-1];

  solver.resetSolver();
  solver.getMoments( ml, mr );
  solver.getRotations( rot );
  printResults();

  while( solver.autoStepSolver() )
  {
   solver.getMoments( ml, mr );
   solver.getRotations( rot );
   printResults();
  }

// Direct solution:
//  solver.goThruSolver();
//  solver.getMoments( ml, mr );
//  solver.getRotations( rot );
//  printResults();
 }
}