/*
 * @(#)CrossSolver.java   1.0 17/12/1999
 */

import java.util.Vector;

/**
 * CrossSolver.java - Solver of Cross Process of continous beams.
 *
 * <p>
 * Description:
 * <p>
 * ===============================================================
 * <p>
 * This file contains an implementation of the Cross Process
 * for continuous beams.  For more details of this process,
 * refer to class notes of course CIV1127-Analise de Estruturas
 * II, Dept. of Civil Engineering, PUC-Rio.
 *
 * <p>
 * ===============================================================
 *
 * @version    1.0 17/12/1999
 * @author     Luiz Fernando Martha
 * @author     Andre Cahn
 *
 */
public class CrossSolver
{

 /*
 ** ---------------------------------------------------------------
 ** Local classes: 
 */

 /**
  * This class defines a node of a continuous beam for the Cross Process.
  */
 protected class Node
 {
  double dl;        /* left moment distribution coefficient */
  double dr;        /* right moment distribution coefficient */
  double tl;        /* left moment carry-over (transmission) factor */
  double tr;        /* right moment carry-over (transmission) factor */
  double rot;       /* node rotation */
 }

 /**
  * This class defines a member of a continuous beam for the Cross Process.
  */
 protected class Member
 {
  double len;       /* length */
  double q;         /* vertical distributed load (top-down positive) */
  double ml;        /* left end moment */
  double mr;        /* right end moment */
  double k;         /* rotational stiffness coefficient */
 }

 /**
  * This class defines, for one step of the Cross Process, a pair 
  * of balancing moments at a node and the carried over moments 
  * to the opposite ends of the adjacent members.
  */
 class BalMoment
 {
  int n;            /* index of node being balanced */
  double bml;       /* balancing moment at left of node */
  double bmr;       /* balancing moment at right of node */
  double tml;       /* carry-over moment at left of node */
  double tmr;       /* carry-over moment at right of node */
 }

 /*
 ** ---------------------------------------------------------------
 ** Protected instance variables:
 */

 /** Number of members. */
 protected int n_member;

 /** Number of nodes. */
 protected int n_node;

 /** Main array for nodes. */
 protected Node node[];

 /** Main array for members. */
 protected Member member[];

 /** Left end support condition. */
 protected int sup_init = 0;

 /** Right end support condition. */
 protected int sup_end = 0;

 /** Tolerance for moment balancing. */
 protected double mom_tol = 0.01;

 /** Flexural stiffness. */
 protected double EI = 10000.0;

 /** Flag to save intermediate results. */
 protected boolean save_steps = true;

 /*
 ** ---------------------------------------------------------------
 ** Package instance variables:
 */

 /** Main vector for balancing moments. */
 Vector balmom;

 /*
 ** ---------------------------------------------------------------
 ** Private and Protected methods: 
 */

 /*=========================  initStiffness  =======================*/
 /**
  * Initialize member stiffness coefficients.
  */
 protected void initStiffness()
 {
  if( sup_init == 0 )
   member[0].k = 3.0 * EI / member[0].len;
  else
   member[0].k = 4.0 * EI / member[0].len;

  if( sup_end == 0 )
   member[n_member-1].k = 3.0 * EI / member[n_member-1].len;
  else
   member[n_member-1].k = 4.0 * EI / member[n_member-1].len;
  
  for( int i=1; i<n_member-1; i++ )
   member[i].k = 4.0 * EI / member[i].len;
 }

 /*===========================  initNodes  =========================*/
 /**
  * Initialize nodes: compute moment distribution coefficients and
  * carry-over factors.
  */
 protected void initNodes()
 {
  double kl;      /* left relative rotational stiffness */
  double kr;      /* right relative rotational stiffness */

  for( int i=0; i<n_node; i++ )
  {
   if( (i == 0) && (sup_init == 0) )
   {
    kl = 3.0 / member[i].len;
    node[i].tl = 0.0;
   }
   else
   {
    kl = 4.0 / member[i].len;
    node[i].tl = 0.5;
   }

   if( (i == n_node-1) && (sup_end == 0) )
   {
    kr = 3.0 / member[i+1].len;
    node[i].tr = 0.0;
   }
   else
   {
    kr = 4.0 / member[i+1].len;
    node[i].tr = 0.5;
   }

   node[i].dl = kl / (kl + kr);
   node[i].dr = kr / (kl + kr);
  }
 }

 /*==========================  initMoments  ========================*/
 /**
  * Initialize member moments with fixed end values.
  */
 protected void initMoments()
 {
  double len;                 /* member length */
  double len2;                /* square of length */
  double q;                   /* member load */

  len = member[0].len;
  len2 = len * len;
  q = member[0].q;
  if( sup_init == 0 )
  {
   member[0].ml = 0.0;
   member[0].mr = - q * len2 / 8.0;
  }
  else
  {
   member[0].ml =   q * len2 / 12.0;
   member[0].mr = - q * len2 / 12.0;
  }

  len = member[n_member-1].len;
  len2 = len * len;
  q = member[n_member-1].q;
  if( sup_end == 0 )
  {
   member[n_member-1].ml = q * len2 / 8.0;
   member[n_member-1].mr = 0.0;
  }
  else
  {
   member[n_member-1].ml =   q * len2 / 12.0;
   member[n_member-1].mr = - q * len2 / 12.0;
  }
  
  for( int i=1; i<n_member-1; i++ )
  {
   len = member[i].len;
   len2 = len * len;
   q = member[i].q;
   member[i].ml =   q * len2 / 12.0;
   member[i].mr = - q * len2 / 12.0;
  }
 }

 /*======================  getMaxUnbalNode  =======================*/
 /**
  * Gets the node of continuous beam that has the maximum absolute
  * value of unbalanced moment.
  * @return the index of the found node.  If all the nodes have 
  * unbalanced moment value below current moment tolerance, 
  * returns a non-valid negative (-1) index.
  */
 protected int getMaxUnbalNode()
 {
  double unbal;      /* unbalanced moment */
  double max_unbal;  /* maximum absolute unbalanced moment */
  int n = -1;        /* index of node with maximum unbalancing */

  max_unbal = mom_tol;
  for( int i=0; i<n_node; i++ )
  {
   unbal = member[i].mr + member[i+1].ml;
   if( Math.abs( unbal ) > max_unbal )
   {
    n = i;
    max_unbal = Math.abs( unbal );
   }
  }

  return( n );
 }

 /*=========================  processNode  =========================*/
 /**
  * Performs one iterative step of the Cross Process of a continuous
  * beam.  Distributes the unbalanced moment of a given node among 
  * the two adjacent beam sections.  Transmits the carry-over 
  * portions of the balancing moments to the opposite ends of the
  * adjacent members.
  * @param n is the index of the target node to be processed
  */
 protected void processNode( int n )
 {
  double unbal;    /* unbalanced moment */
  double bml;      /* balancing moment at left of node */
  double bmr;      /* balancing moment at right of node */
  double tml;      /* carry-over moment at left of node */
  double tmr;      /* carry-over moment at right of node */
  double drot;     /* node rotation increment */

  unbal = member[n].mr + member[n+1].ml;
  bml = - unbal * node[n].dl;
  bmr = - unbal * node[n].dr;
  tml = bml * node[n].tl;
  tmr = bmr * node[n].tr;
  member[n].ml   += tml;
  member[n].mr   += bml;
  member[n+1].ml += bmr;
  member[n+1].mr += tmr;

  if( save_steps )
  {
   BalMoment bm = new BalMoment();
   bm.n = n;
   bm.bml = bml;
   bm.bmr = bmr;
   bm.tml = tml;
   bm.tmr = tmr;
   balmom.add( bm );
  }

  drot = - unbal / (member[n].k + member[n+1].k);
  node[n].rot += drot;
 }

 /*
 ** ---------------------------------------------------------------
 ** Public methods: 
 */

 /*=======================  Constructor  =========================*/
 /**
  * Creates a CrossSolver object for a complete set of data.
  */
 public CrossSolver( int n_member, int sup_init, int sup_end, 
                     double mom_tol, double len[], double q[],
                     double EI )
 {
  this.n_member = n_member;
  n_node = n_member - 1;
  this.sup_init = sup_init;
  this.sup_end = sup_end;
  this.mom_tol = mom_tol;
  this.EI = EI;

  member = new Member[n_member];
  for( int i=0; i<n_member; i++ )
  {
   member[i] = new Member();
   member[i].len = len[i];
   member[i].q = q[i];
  }

  node = new Node[n_node];
  for( int j=0; j<n_node; j++ )
  {
   node[j] = new Node();
  }

  balmom = new Vector();

  initStiffness();
  initNodes();
 }

 /*=======================  Constructor  =========================*/
 /**
  * Creates a CrossSolver object for a complete set of data, with
  * exception of the flexural stiffness that uses a default value.
  */
 public CrossSolver( int n_member, int sup_init, int sup_end, 
                     double mom_tol, double len[], double q[] )
 {
  this( n_member, sup_init, sup_end, mom_tol, len, q, 10000.0 );
 }

 /*=======================  Constructor  =========================*/
 /**
  * Creates a default CrossSolver object.
  */
 public CrossSolver()
 {
  n_member = 3;
  n_node = n_member - 1;
  mom_tol = 0.01;
  sup_init = 0;
  sup_end = 1;
  EI = 10000.0;

  member = new Member[n_member];
  member[0] = new Member();
  member[0].len = 8.0;
  member[0].q = 6.0;
  member[1] = new Member();
  member[1].len = 6.0;
  member[1].q = 6.0;
  member[2] = new Member();
  member[2].len = 8.0;
  member[2].q = 6.0;

  node = new Node[n_node];
  for( int j=0; j<n_node; j++ )
  {
   node[j] = new Node();
  }

  balmom = new Vector();

  initStiffness();
  initNodes();
 }

 /*=======================  setMomentToler  =========================*/
 /**
  * Set moment balancing tolerance.
  */
 public void setMomentToler( double mom_tol )
 {
  this.mom_tol = mom_tol;
 }

 /*=========================  resetSolver  =========================*/
 /**
  * Initializes solver of Cross Process for continuous beams.
  * This method sets up the CrossSolver object for a new solution.
  */
 public void resetSolver()
 {
  initMoments();

  for( int i=0; i<n_node; i++ )
   node[i].rot = 0.0;
 
  balmom.removeAllElements();
 }

 /*=======================  nodeStepSolver  ========================*/
 /**
  * Solver one step of Cross Process for continuous beams for a given
  * node.
  */
 public boolean nodeStepSolver( int n )
 {
  double unbal;      /* unbalanced moment */

  if( (n < 0) || (n >= n_node) )
   return( false );

  unbal = member[n].mr + member[n+1].ml;
  if( Math.abs( unbal ) < mom_tol )
   return( false );

  processNode( n );

  return( true );
 }

 /*========================  moreStepsToGo  ========================*/
 /**
  * Check to see whether there is more Cross steps to go: returns a
  * true status if there is at least one step to go; returns a false
  * status otherwise.
  */
 public boolean moreStepsToGo()
 {
  int n;
  
  n = getMaxUnbalNode();
  if( n < 0 )
   return( false );

  return( true );
 }

 /*=======================  autoStepSolver  ========================*/
 /**
  * Solver one step of Cross Process for continuous beams: solve the
  * node with maximum absolute value of unbalanced moment.
  */
 public boolean autoStepSolver()
 {
  int n;
  
  n = getMaxUnbalNode();
  if( n < 0 )
   return( false );

  processNode( n );

  return( true );
 }

 /*========================  goThruSolver  =========================*/
 /**
  * Direct solver of Cross Process for continuous beams.
  */
 public void goThruSolver()
 {
  resetSolver();

  while( autoStepSolver() );
 }

 /*========================  getNumMembers  =========================*/
 /**
  * Get number of member of this object.
  */
 public int getNumMembers()
 {
  return( n_member );
 }

 /*========================  getMoments  =========================*/
 /**
  * Get current member moments of this CrossSolver object.
  */
 public void getMoments( double ml[], double mr[] )
 {
  for( int i=0; i<n_member; i++ )
  {
   ml[i] = member[i].ml;
   mr[i] = member[i].mr;
  }
 }

 /*=======================  getRotations  ========================*/
 /**
  * Get current node rotations of this CrossSolver object.
  */
 public void getRotations( double rot[] )
 {
  for( int i=0; i<n_node; i++ )
  {
   rot[i] = node[i].rot;
  }
 }
}